class Joypad:
    """Treated as a custom type for joypad layouts."""
    pass


class NESJoypad(Joypad):
    """Controller mapping for the NES.
    
    The \"Power\" and \"Reset\" buttons are not implemented for the time being."""
    def __init__(self):
        super().__init__()
        self._name = "NES"
        self.controls = {
            "A": False, "B": False, "Up": False, "Down": False, "Right": False, "Left": False,
            "Select": False, "Start": False,
        }


class GBAndGBCJoypad(Joypad):
    """Controller mapping for the GB and GBC.
    
    The \"Power\" button is not implemented for the time being."""
    def __init__(self):
        super().__init__()
        self._name = "GB(C)"
        self.controls = {
            "A": False, "B": False, "Up": False, "Down": False, "Right": False, "Left": False,
            "Select": False, "Start": False,
        }


class SNESJoypad(Joypad):
    """Controller mapping for the SNES.
    
    The \"Power\" and \"Reset\" buttons are not implemented for the time being."""
    def __init__(self):
        super().__init__()
        self._name = "SNES"
        self.controls = {
            "A": False, "B": False, "L": False, "R": False, "X": False, "Y": False, "Up": False,
            "Down": False, "Right": False, "Left": False, "Select": False, "Start": False,
        }


class Memory:
    """Class containing various functions for setting controller input in BizHawk."""
    def __init__(self, client) -> None:
        self.client = client
        self.address_list = []
        self.received_memory = {}

    def add_address(self, addr):
        """Adds an address for the external tool to return.

        :param addr: A hexidecimal value representing the address to read from \
        in the BizHawk emulator's memory.
        :type addr: int"""
        if not str(addr) in self.address_list:
            self.address_list.append(str(addr))
            self.received_memory[hex(addr)] = -1

    def add_address_range(self, start, end):
        """Adds a range of addresses from `start` to `end`, both inclusive.

        :param start: A hexidecimal value representing the first address in the range.
        :type start: int

        :param end: A hexidecimal value representing the last address in the range.
        :type end: int

        :precondition: `start` <= `end`."""
        for addr in range(int(start), int(end) + 1):
            self.add_address(addr)

    def request_memory(self):
        """Requests for the latest memory data from the external tool."""
        if self.address_list != []:
            self.client._queue_request("MEM_ADDRESS;" + ";".join(self.address_list) + "\n")
            self.address_list = []
        self.client._queue_request("MEM_READ;\n")

    def get_memory(self) -> dict:
        """Gets the latest memory data received from the external tool. 

        This will return a copy of the dictionary containing the latest data
        received from each requested address. Where the address
        (in hexadecimal representation) is the key, and the data is the value
        (in decimal representation).

        The value is set to `-1` if no data has been received for that address."""
        data = self.client._get_latest_response_data("MEM_READ")
        if data:
            address_value_pairs = data.strip(";").split(";")
            for addr_val_pair in address_value_pairs:
                temp = addr_val_pair.split(":")
                addr, val = temp[0], temp[1]
                self.received_memory[hex(int(addr))] = int(val)
        return self.received_memory.copy()



class ExternalTool:
    """Class containing various functions for controlling the external tool."""
    def __init__(self, client) -> None:
        self.client = client

    def request_commandeer(self):
        """Requests for the current status on commandeer."""
        self.client._queue_request("COM_GET;\n")

    def get_commandeer(self) -> bool:
        """Returns the current status on commandeer: `True` if enabled, `False` otherwise."""
        data = self.client._get_latest_response_data("COM_GET")
        return eval(data)

    def set_commandeer(self, enabled):
        """Sets the communication mode of the external tool to either manual or commandeer.

        :param enabled: Determines whether a request to enable commandeer or disable it should be sent. \
        `True` = enable commandeer, `False` = disable.
        :type enabled: bool"""
        self.client._queue_request("COM_SET;" + str(enabled) + "\n")



import socket

from birdseyelib.response import Response


class Client:
    """A socket client used to communicate with the external tool 
    located at a set port and address.

    Use the external tool to get and set the port and
    address the server should listen in to.

    :param ip: The socket address.
    :type ip: str

    :param port: The socket port number.
    :type port: int"""
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.client = None  # Initialized in Client.connect
        self.connection_status = -1
        self.payload = ""
        self.latest_responses = []

    def connect(self):
        """(Blocking) Attempt to connect to the external tool.

        Call `Client.is_connected()` to see if the attempt was successful."""
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection_status = self.client.connect_ex((self.ip, self.port))

    def close(self):
        """Close socket connection with external tool and send a final message
        notifying the external tool before closing."""
        try:
            self.client.sendall("CLOSE;\n".encode())
        except:
            pass
        self.client.close()
        self.connection_status = -1

    def is_connected(self):
        """Returns true if client is connected to the external tool."""
        return self.connection_status == 0

    def _send_requests(self):
        """(Internal) Send requests from any birdseye object to the external
        tool.

        :precondition: client is connected to a socket."""
        try:
            self.client.sendall(self.payload.encode())
            self.payload = ""
        except:
            self.close()

    def _queue_request(self, request):
        """(Internal) Append a string request to the queue to be
        sent.
        
        :param request: The request to be added.
        :type request: str"""
        self.payload += request

    def _receive_messages(self) -> str:
        """(Internal) (Blocking) Wait to receive data collected by the external
        tool.

        :precondition: client is connected to a socket."""
        message = ""
        bufsize = 2048
        try:
            message = self.client.recv(bufsize).decode()
            if (len(message) >= bufsize):  # Full message is possibly larger than `bufsize`
                self.client.setblocking(False)
                try:
                    temp = self.client.recv(bufsize).decode()
                    while (True):
                        message += temp
                        temp = self.client.recv(bufsize).decode()
                except:  # `socket.recv` will raise an exception if empty.
                    pass
                self.client.setblocking(True)
            return message
        except InterruptedError:
            self.close()

    def _parse_responses(self):
        """(Internal) (Blocking) Parse received responses into a list of
        objects.
        
        :precondition: client is connected to a socket."""
        data = self._receive_messages()
        if data:
            responses = data.strip("\n").split("\n")
            self.latest_responses.clear()
            for response in responses:
                self.latest_responses.append(Response(*response.split(";", maxsplit=1)))

    def _get_latest_response_data(self, tag) -> str:
        """(Internal) Returns the latest response received with the given
        `tag`. An empty string is returned if a response with `tag` cannot be
        found.
        
        :param tag: The tag corresponding with the received data of interest.
        :type tag: str"""
        for response in self.latest_responses:
            if tag == response.tag:
                return response.data
        
        return ""

    def advance_frame(self):
        """(Blocking) Sends all queued requests to the external tool and parses
        received responses. This will subsequently advance the emulator to the
        next frame.
        
        :precondition: client is connected to a socket."""
        self._send_requests()
        self._parse_responses()



class ControllerInput:
    """Class containing various functions for reading memory from BizHawk."""
    def __init__(self, client):
        self.client = client

    def set_joypad(self, joypad):
        """Sets the joypad layout to be used when setting inputs.
        
        Layout is set to the NES joypad by default.

        :param joypad: The joypad layout to send to the external tool. \
        Make sure the joypad matches the emulated system being used!
        :type joypad: Joypad"""
        self.client._queue_request("INP_JOYPAD;" + joypad._name + "\n")

    def set_controller_input(self, joypad):
        """Sets the controller inputs to be executed in the emulator.
        All inputs are set to `False` be default.
        The inputs are executed until a new controller input is sent.

        :param joypad: The joypad layout to send to the external tool. \
        Make sure the joypad matches the emulated system being used!
        :type joypad: Joypad"""
        bool_to_string = {False : "false", True : "true"}
        controller_input = ";".join(
            [bool_to_string[joypad.controls[button]] for button in joypad.controls.keys()]
        )
        self.client._queue_request("INP_SET;" + controller_input + "\n")



class Emulation:
    """Class containing various functions for controlling emulation in BizHawk."""
    def __init__(self, client) -> None:
        self.client = client

    def request_framecount(self):
        """Requests for the current framecount from the external tool."""
        self.client._queue_request("EMU_FRAME;\n")
    
    def request_board_name(self):
        """Requests for the board name of the loaded ROM."""
        self.client._queue_request("EMU_BOARD;\n")

    def get_framecount(self) -> int:
        """Returns the current framecount from the emulator, or `-1`, if no
        data has been received yet."""
        data = self.client._get_latest_response_data("EMU_FRAME")
        return int(data) if data else -1
    
    def get_board_name(self) -> str:
        """Returns the board name of the loaded ROM, or `""`, if not available."""
        data = self.client._get_latest_response_data("EMU_BOARD")
        return data if data else ""
